% script per caricare i dati da "furnace.dat", creare delle strutture dati
% per il SysId toolbox e disegnare i grafici
%
% origine dei dati: Time Series Data Library
% URL: https://datamarket.com/data/set/2372/methane-input-into-gas-furnace-cu-ftmin-sampling-interval-9-seconds-carbon-dioxide-output-from-gas-furnace-percent-of-output-gas-sampling-interval-9-seconds#!ds=2372!2juz&display=line
% 
% dati tratti da:
% Box & Jenkins, "Time Series Analysis: Forecasting and Control",
% Holden-Day 1976
%
% NB utilizza il toolbox --> Generate maximally perceptually-distinct colors, version 1.2.0.0 by Tim Holy
% scaricabile da: https://it.mathworks.com/matlabcentral/fileexchange/29702-generate-maximally-perceptually-distinct-colors

clear all
close all hidden
clc

%% -------- carico i dati e disegno

%-------------------------------------------------------------------------
% apro il file 
fId = fopen('furnace.dat');
data_headers = textscan(fId, '%s', 11, 'delimiter', '\n');

eia_series1_headers = data_headers{1}; % e memorizzo l'intestazione


%-------------------------------------------------------------------------
% carico i dati
% formato dei dati:
% "\" dato numerico "spazio" dato numerico (296 coppie di valori)
%
data = textscan(fId, '\\%f %f', 296);
CO2_percent_data = data{1,1}; % percentuale di CO2 nei gas in uscita: Y
CH4_data = data{1,2}; % metano in ingresso nella fornace; cu. ft/min; U

% chiudo il file
fclose(fId);

%-------------------------------------------------------------------------
U = CH4_data;
Y = CO2_percent_data;
Ts = 9.0; % campionamento
time_vect = (0: numel(U)-1)* Ts; % istanti di tempo associati ai campioni

mpdc2 = distinguishable_colors(2); % 2x3 color list
plotLineWidth = 0.5;
plotMarkerSize = 6;

hf=figure('Name','B&J furnace - samples');
axhnd1 = subplot(2,1,1);
plot(time_vect, U, '-s', 'LineWidth', plotLineWidth, ...
    'Color', mpdc2(1,:) , 'MarkerSize', plotMarkerSize, ...
    'MarkerEdgeColor', mpdc2(1,:),...
    'MarkerFaceColor', mpdc2(1,:)); grid on;
xlabel('time instants [s]'); 
ylabel(sprintf('Methane input into gas furnace \n[ft^3/min]'));
axis tight
axhnd2 = subplot(2,1,2);
plot(time_vect, Y,'-s', 'LineWidth', plotLineWidth , ...
    'Color', mpdc2(2,:), 'MarkerSize', plotMarkerSize, ...
    'MarkerEdgeColor', mpdc2(2,:),...
    'MarkerFaceColor', mpdc2(2,:)); grid on;
xlabel('time instants [s]'); 
ylabel(sprintf('Carbon dioxide output \n from gas furnace \n[percent of output gas]'));
axis tight
linkaxes([axhnd1 axhnd2],'x'); % cosi' sincronizzo lo zoom nei 2 grafici

%% ---------- pre-elaborazione dei dati: detrend
data_set = iddata(Y, U, Ts, 'OutputName','diff. Co_2 in %',...
    'OutputUnit','%', 'InputName','diff portata di metano',...
    'InputUnit','cu. ft/min','Tstart',0,'TimeUnit','seconds'); 
% oggetto iddata che memorizza tutti i dati
[data_s_d T] = detrend(data_set); 
% rendo i dati a media nulla e tolgo eventuali trend

%======================================================================
% i dati non sono abbastanza numerosi da poter applicare cross-validation
% per scegliere l'ordine del modello: lo determino usando FPE, MDL, AIC

% ordine di persistente eccitazione?
pe_n = pexcit(data_s_d);
messaggio = sprintf('L''ordine di persistente eccitazione vale almeno n= %d ', pe_n);
disp('************************************************************');
disp(messaggio);
disp('************************************************************');

pause(3); 
disp('--------------------------------------------------');
disp('NB: l''elaborazione puo'' durare anche 4-5 minuti!');
disp('--------------------------------------------------');

% per avere AIC, FPE, MDL procedo in questo modo: identifico tutti i
% modelli usando tutti i dati a disposizione e poi valuto gli indici,
% memorizzando il risultato per ciascun modello. Infine, esplorando le
% curve degli indici, determino i "vincitori"

complessitaARX = struc(1:5, 1:5, 1:10);
% voglio esplorare gli ARX per n_a da 1 a 15, n_b da 1 a 10, n_k da 1 a 10

% inizializzo i vettori dove memorizzare gli indici
AIC_ARXscores = zeros(size(complessitaARX,1),1);
FPE_ARXscores = zeros(size(AIC_ARXscores));
MDL_ARXscores = zeros(size(AIC_ARXscores));


% e la struttura che memorizza i modelli ARX
ARX_set = cell(size(AIC_ARXscores));


%% -------- analisi nella famiglia ARX
tic
for ii=1: size(complessitaARX,1)
    modello_ARX = arx(data_s_d, complessitaARX(ii,:));
    ARX_set{ii} = modello_ARX;
    % identifico un modello ARX di ordine imposto dal vettore riga
    % complessitaARX(ii,:)
    % ATTENZIONE: codice NON efficiente
    % impiega diverso tempo ad eseguire le valutazioni richieste (piu' di 6
    % minuti su un Core 2 Duo a 3 GHz)
    AIC_ARXscores(ii) = aic(modello_ARX);
    FPE_ARXscores(ii) = fpe(modello_ARX);
    MDL_ARXscores(ii) = mdl_score(modello_ARX);
    % disp(ii);
end % for ii
tempo_speso = toc;
fprintf(1, 'tempo speso per l''analisi dei modelli ARX %f [s]\n', tempo_speso);

[aic_min,aic_id] = min(AIC_ARXscores);
[fpe_min,fpe_id] = min(FPE_ARXscores);
[mdl_min,mdl_id] = min(MDL_ARXscores);


aic_best_arx = ARX_set{aic_id};aic_best_arx.Name = 'ARX model with best AIC score';
fpe_best_arx = ARX_set{fpe_id};fpe_best_arx.Name = 'ARX model with best FPE score';
mdl_best_arx = ARX_set{mdl_id};mdl_best_arx.Name = 'ARX model with best MDL score';

disp('===================================');
disp('il miglior modello ARX secondo AIC:');
disp(aic_best_arx);
disp('------------');
fprintf(1,...
    'complessita'' del modello: n_a=%g, n_b=%g, n_k=%g \n',...
    complessitaARX(aic_id,:));
disp('===================================');
disp('il miglior modello ARX secondo FPE:');
disp(fpe_best_arx);
disp('------------');
fprintf(1,...
    'complessita'' del modello: n_a=%g, n_b=%g, n_k=%g \n',...
    complessitaARX(fpe_id,:));

disp('===================================');
disp('il miglior modello ARX secondo MDL:');
disp(mdl_best_arx);disp('------------');
fprintf(1,...
    'complessita'' del modello: n_a=%g, n_b=%g, n_k=%g \n',...
    complessitaARX(mdl_id,:));

disp('=====================================');


% ed ora i test sui residui
resOpt = residOptions('Maxlag', floor(numel(U)/4));
% valuta la funzione di autocorrelazione tra -N/4 e +N/4, dove N numero di
% campioni
hfc = figure;resid(data_s_d, aic_best_arx, resOpt);
set(hfc,'Name',[get(hfc,'Name'),' - model:', aic_best_arx.Name])


hfc = figure;resid(data_s_d, fpe_best_arx, resOpt);
set(hfc,'Name',[get(hfc,'Name'),' - model:', fpe_best_arx.Name])

hfc = figure;resid(data_s_d, mdl_best_arx, resOpt);
set(hfc,'Name',[get(hfc,'Name'),' - model:', mdl_best_arx.Name])

% -----------------------------
% ed ora i test di Anderson sui residui

% ed ora i test di Anderson sui residui
H_AIC = AndersonWhitenessTest(aic_best_arx, data_s_d, 0.01, true);
% test: si puo' negare che il residuo di predizione sia 
%       realizzazione di processo bianco?
% se H = 0 -> non si puo' negare
% se H = 1 -> ipotesi rigettata: la distribuzione NON e' quella desiderata,
%               i residui NON sono di un processo di rumore bianco

H_FPE = AndersonWhitenessTest(fpe_best_arx, data_s_d, 0.01, true);

H_MDL = AndersonWhitenessTest(mdl_best_arx, data_s_d, 0.01, true);

% NB quale modello usare?

% e le prestazioni?
% come predittori ad 1 passo, quindi ciascuno "alimentato" ad
% ogni passo dalle misure di ingresso ed uscita fino al passo precedente
figure;compare(data_s_d, aic_best_arx,fpe_best_arx, mdl_best_arx, 1 );
fh = gcf; set(fh,'Name','predizione');

%% ===========================codice ARX efficiente============
% togliere il commento a questa porzione di codice e confrontare il
% risultato in termini di tempo e soluzioni proposte
%{
tic
V_arx = arxstruc(data_s_d, data_s_d, complessitaARX);
tempo_2 = toc;
% NB usando il medesimo set di dati per identificazione e validazione
% OVVIAMENTE una delle proposte successive e' sbagliata (il "best fit" che
% minimizza l'errore di validazione in realta' non e' da considerare
auto_arx_bestAIC = selstruc(V_arx, 'aic');
auto_arx_best_MDL = selstruc(V_arx, 'mdl');

fprintf(1,...
    'analisi dei modelli ARX - algoritmo efficiente - %f [s]\n',...
    tempo_2);

disp('***********************************');
disp('il miglior modello ARX secondo AIC:');
disp(auto_arx_bestAIC);
disp('***********************************');
disp('il miglior modello ARX secondo MDL:');
disp(auto_arx_best_MDL);
disp('***********************************');

%}


